# -*- coding: utf-8 -*-
"""
# @Time    : 2022/6/9 19:30
# @Author  : bruce
# @desc    :
"""


def singleton(cls):
    _instance = {}

    def inner(*args, **kwargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kwargs)
        return _instance[cls]
    return inner



# -*- coding: utf-8 -*-
"""
# @Time    : 2022/3/4 15:13
# @Author  : bruce
# @desc    :
"""
import copy
import json
import os

import tablib
import ast
from ddt import ddt, data, feed_data

from api_test_ez.core.case.frame.frame_unittest import UnitHttpFrame
from api_test_ez.core.case.http.request import Request
from api_test_ez.core.case.http.response import EzResponse

from api_test_ez.ez import Http
from api_test_ez.project import Project


def load_test_data(data_filename):
    if isinstance(data_filename, str):
        with open(data_filename, 'rb') as f:
            data_set = tablib.Dataset().load(f.read())
            return data_set.dict
    else:
        return []


def ez_ddt_setter(cls):
    """Set `%values` (`DATA_ATTR` from ddt) for testMethods.

    `ddt` will copy methods which have `DATA_ATTR` attribute.
    So we set it to all the testMethods (sometimes we need set some method exactly,
    it will be supported maybe next version).
    Then Unittest can run all of these cases."""
    DATA_ATTR = "%values"
    if hasattr(cls, "load_data"):
        load_func = getattr(cls, "load_data")
        values = getattr(load_func, DATA_ATTR)
        for name, func in list(cls.__dict__.items()):
            if name.startswith("test"):
                setattr(func, DATA_ATTR, values)
    return cls


DATA_HOLDER = "%data_holder"


class CaseMetaclass(type):
    """Mapping test method and ddt data.
    `ddt` only copy the method which is decorated by `data`,
    let's make test_method and ddt_data map."""
    def __new__(mcs, name, bases, attrs):
        super_new = super().__new__

        # If a base class just call super new
        if name == 'UnitCase':
            return super_new(mcs, name, bases, attrs)

        for base_class in bases:

            if base_class is UnitCase:
                # Find data functions
                ddt_func_names = []
                for base_attr_name in dir(base_class):
                    if base_attr_name.startswith('load_data'):
                        ddt_func_names.append(base_attr_name)

                # If data function is None, let test run itself.
                if len(ddt_func_names) == 0:
                    return super_new(mcs, name, bases, attrs)

                new_attrs = {}
                # Mapping test methods and data functions
                for name, func in attrs.items():
                    if name.startswith('test'):
                        for ddt_func_name in ddt_func_names:
                            test_name = ddt_func_name.replace('load_data', name)
                            # Let's set a `%data_owner` attr to test function
                            # Then we can find the data later.

                            # We can not copy function directly, create it from `ddt.feed_data`.
                            _func = feed_data(func, test_name, func.__doc__)
                            setattr(_func, DATA_HOLDER, ddt_func_name)
                            new_attrs.update({test_name: _func})
                    else:
                        # other methods should be added without modified
                        new_attrs.update({name: func})

                return super_new(mcs, name, bases, new_attrs)

        return super_new(mcs, name, bases, attrs)


@ddt
class UnitCase(UnitHttpFrame, metaclass=CaseMetaclass):
    # env init
    case_path_dir = os.getcwd()
    ez_project = Project(ez_file_path=case_path_dir, env_name=os.path.basename(case_path_dir))
    configs = ez_project.configs
    logger = ez_project.logger
    # load test data
    __casefile__ = configs.get("case_filepath")
    data_set = load_test_data(__casefile__)
    # set request here, bcz the data in `ez.config` can not be load in again.

    __autoRequest__ = configs.get("auto_request")

    def __new__(cls, methodName, *args, **kwargs):
        # bcz of `__classcell__` error, copy config at here.
        cls.local_config = copy.deepcopy(cls.configs)
        return super(UnitHttpFrame, cls).__new__(cls, *args, **kwargs)

    def __init__(self, methodName):
        self.request = Request(http=Http())
        self.response = EzResponse(logger=self.logger)
        self.request.owner = methodName
        self.response.owner = methodName
        self.initRequest(methodName)
        super(UnitCase, self).__init__(methodName)

    @data(*data_set)
    def load_data(self, case_data: dict):
        for key, value in case_data.items():
            self.local_config.set(key, value, priority="case")

    def initRequest(self, testmethod_name):
        # Find my ddt data holder via testmethod_name.
        if hasattr(self, testmethod_name):
            test_func = getattr(self, testmethod_name)
            if hasattr(test_func, DATA_HOLDER):
                data_holder = getattr(test_func, DATA_HOLDER)
                if hasattr(self, data_holder):
                    _ddt_data_func = getattr(self, data_holder)
                    _ddt_data_func()
        self.request.set(self.local_config)
        return self.request

    def doRequest(self, request=None):
        if request:
            self.request.set(request)
        self.logger.debug(repr(self.request))
        # Prepare request
        http = self.request.http
        url = self.request.url
        method = self.request.method.lower()

        body = self.request.body
        body_type = self.request.body_type

        # Request start
        if url and hasattr(http, method):
            do = getattr(http, method)
            self.response.set(do(url=url, **{body_type: body}))
            self.logger.debug(repr(self.response))



# -*- coding: utf-8 -*-
"""
# @Time    : 2022/5/12 17:54
# @Author  : bruce
# @desc    :
"""
import logging
from json import JSONDecodeError

from requests import Response

from api_test_ez.ez.decorator.jsonbean import json_bean
from api_test_ez.ez.serialize.errors import ValidationError
from api_test_ez.ez.serialize.models import ValidatorModel
from marshmallow import Schema

from api_test_ez.project import get_ez_logger, get_ez_settings


class EzResponse(Response):

    __slots__ = ("owner", "response", "logger")

    def __init__(self, logger, response: Response = None):
        super().__init__()
        self.response = response
        self.logger = logger

    # def __getattribute__(self, item):
    #     if item not in ("owner", "response", "logger") and not item.startswith('__') and hasattr(self.response, item):
    #         return self.response.__getattribute__(item)
    #     else:
    #         return super(EzResponse, self).__getattribute__(item)

    def __getattr__(self, item):
        return self.response.__getattribute__(item)

    def set(self, response: Response):
        self.response = response

    @property
    @json_bean
    def bean(self):
        return self.json()

    def pair(self, model: ValidatorModel, full_repr=False):
        validate_result = model.validate(self.response.json(), full_repr)
        if 'ValidationError' in str(validate_result):
            validation_error = ValidationError(f'[{self.owner}] {validate_result}')
            self.logger.error(validation_error)
            raise validation_error

    def validate(self, schema: Schema):
        """Validate from marshmallow."""
        return schema.load(self.response.json())

    def __str__(self):
        if self.response:
            return f"<{self.__class__.__name__}> {self.owner}:\n" \
                   f"{self.response.text!r}"
        else:
            return f"<{self.__class__.__name__}> {self.owner}: response is None, " \
                   f"maybe you didn't send the request?"

    __repr__ = __str__



# -*- coding: utf-8 -*-
"""
# @Time    : 2022/3/5 17:00
# @Author  : bruce
# @desc    :
"""
from configparser import ConfigParser


def get_config(file_path):
    """Get Ez config file as a ConfigParser"""
    cfg = ConfigParser()
    cfg.read(file_path, encoding='utf-8')
    return cfg



# -*- coding: utf-8 -*-
"""
# @Time    : 2022/3/3 15:24
# @Author  : bruce
# @desc    :
"""
from enum import Enum

import requests
import urllib3
urllib3.disable_warnings()


class Method:
    GET = 'GET'
    POST = 'POST'
    HEAD = 'HEAD'
    OPTIONS = 'OPTIONS'
    DELETE = 'DELETE'
    PUT = 'PUT'
    PATCH = 'PATCH'


class Http:

    def __init__(self, headers=None, timeout=3, allow_redirects=False, verify=False, proxies=None, retry=1):
        """
        Http初始化
        :param headers: 请求头。
        :param timeout: 超时时间。
        :param with_session: 是否创建session，默认创建。
        :param allow_redirects: 是否允许重定向，默认否。
        :param verify: 是否进行https证书验证，默认否。
                       如果开启：verify='/<path>/<file_name>.pem'。
        :param retry: 请求重试次数。
        """
        self._headers = headers
        self._timeout = timeout
        self._session = requests.session()
        self._allow_redirects = allow_redirects
        self._verify = verify
        self._proxies = proxies
        self._retry = retry
        self._cookies = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def timeout(self):
        """
        获取超时时间
        :return:
        """
        return self._timeout

    @timeout.setter
    def timeout(self, value: int):
        """
        设置超时时间
        :param value:
        :return:
        """
        self._timeout = value

    @property
    def headers(self):
        """
        获取请求头
        :return:
        """
        return self._headers

    @headers.setter
    def headers(self, value):
        """
        设置请求头
        :param value:
        :return:
        """
        self._headers = value

    @property
    def session(self):
        """
        获取session
        :param value:
        :return:
        """
        return self._session

    @session.setter
    def session(self, value):
        """
        设置session
        :param value:
        :return:
        """
        self._session = value

    @property
    def cookies(self):
        """
        :return:
        """
        return self._cookies

    @cookies.setter
    def cookies(self, value):
        """
        :return:
        """
        self._cookies = value

    @property
    def retry(self):
        """
        获取重试次数
        :return:
        """
        return self._retry

    @retry.setter
    def retry(self, value: int):
        """
        设置重试次数
        :return:
        """
        self._retry = value

    @property
    def proxies(self):
        """
        获取代理
        :return:
        """
        return self._proxies

    @proxies.setter
    def proxies(self, value):
        """
        设置代理
        :param value:
        :return:
        """
        self._proxies = value

    @property
    def allow_redirects(self):
        """
        获取是否允许重定向
        :return:
        """
        return self._allow_redirects

    @allow_redirects.setter
    def allow_redirects(self, value):
        """
        设置是否允许重定向
        :param value:
        :return:
        """
        self._allow_redirects = value

    @property
    def verify(self):
        """
        获取是否开启https认证
        :return:
        """
        return self._verify

    @verify.setter
    def verify(self, value):
        """
        设置是否开启https认证
        :param value:
        :return:
        """
        self._verify = value

    def close(self):
        if self._session:
            self._session.close()

    def m_request(self, method, url, **kwargs):
        for i in range(self._retry):
            try:
                return self.session.request(method=method, url=url, headers=self._headers, cookies=self._cookies,
                                            timeout=self._timeout, allow_redirects=self._allow_redirects,
                                            proxies=self._proxies, verify=self._verify, **kwargs)
            except Exception as e:
                print(e)
        else:
            return None

    def get(self, url, **kwargs):
        """发送 GET 请求. Returns :class:`Response` 对象.

        :param url: URL for the new :class:`Request` object.
        :param **kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """
        return self.m_request(method=Method.GET, url=url, **kwargs)

    def post(self, url, data=None, json=None, **kwargs):
        """发送 POST 请求. Returns :class:`Response` 对象.

        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """
        return self.m_request(method=Method.POST, url=url, data=data, json=json, **kwargs)

    def options(self, url, **kwargs):
        """发送 OPTIONS 请求. Returns :class:`Response` 对象.

        :param url: URL for the new :class:`Request` object.
        :param kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        kwargs.setdefault('allow_redirects', True)
        return self.m_request(method=Method.OPTIONS, url=url, **kwargs)

    def head(self, url, **kwargs):
        """发送 HEAD 请求. Returns :class:`Response` 对象.

        :param url: URL for the new :class:`Request` object.
        :param kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        kwargs.setdefault('allow_redirects', False)
        return self.m_request(method=Method.HEAD, url=url, **kwargs)

    def put(self, url, data=None, **kwargs):
        """发送 PUT 请求. Returns :class:`Response` 对象.

        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        return self.m_request(method=Method.PUT, url=url, data=data, **kwargs)

    def patch(self, url, data=None, **kwargs):
        """发送 PATCH 请求. Returns :class:`Response` 对象.

        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        return self.m_request(method=Method.PATCH, url=url, data=data, **kwargs)

    def delete(self, url, **kwargs):
        """发送 DELETE 请求. Returns :class:`Response` 对象.

        :param url: URL for the new :class:`Request` object.
        :param kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        return self.m_request(method=Method.DELETE, url=url, **kwargs)



# -*- coding: utf-8 -*-
"""
# @Time    : 2022/6/9 19:30
# @Author  : bruce
# @desc    :
"""
import typing
from marshmallow import Schema, ValidationError


class EzSchema(Schema):

    def __getattr__(self, item):
        """Return fields in `_declared_fields`, to dynamically modify field properties."""
        if item in self._declared_fields.keys():
            return self._declared_fields[item]
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")

    def handle_error(
        self, error: ValidationError, data: typing.Any, *, many: bool, **kwargs
    ):
        err_msg = f"Error Message: {error.messages}.\n" \
                  f"Raw Data: {error.data}"
        raise ValidationError(err_msg)



# -*- coding: utf-8 -*-
"""
# @Time    : 2022/6/9 19:30
# @Author  : bruce
# @desc    :
"""
import os
from abc import ABC, abstractmethod

from BeautifulReport import BeautifulReport

from api_test_ez.libs import HTMLTestRunner
from api_test_ez.project import Project
from unittest.loader import defaultTestLoader


class Reporter(ABC):

    def __init__(self, case_path, report_title=None, report_desc=None, *args, **kwargs):
        self.case_file_name = None
        if os.path.isdir(case_path):
            self.case_file_dir = case_path
        else:
            self.case_file_dir = os.path.dirname(case_path)
            self.case_file_name = os.path.basename(case_path)

        project = Project(ez_file_path=self.case_file_dir)
        self.settings = project.settings
        # log
        self.logger = project.logger
        # report file
        self.report = project.report
        self.report_dir = self.report.report_dir
        self.report_file_name = self.report.report_file_name
        self.report_file_path = project.report.report_file_path
        self.report_title = report_title if report_title else project.report.report_title
        self.report_desc = report_desc if report_desc else project.report.report_desc

    def load_tests(self):
        """Load tests from a dir. It is same as the `load_tests` protocol in `unittest`."""
        if self.case_file_name:
            print(self.case_file_name)
            return defaultTestLoader.discover(self.case_file_dir, pattern=f'{self.case_file_name}')
        else:
            return defaultTestLoader.discover(self.case_file_dir, pattern='*.py')

    @abstractmethod
    def run(self):
        pass

    def send(self):
        pass


class HtmlReporter(Reporter):
    """HTMLTestRunner"""

    def run(self):
        if not os.path.exists(self.report_dir):
            os.makedirs(self.report_dir)
        fp = open(self.report_file_path, 'wb')
        try:
            suit = self.load_tests()
            self.logger.info("********TEST BEGIN********")
            runner = HTMLTestRunner.HTMLTestRunner(stream=fp, title=f'{self.report_title}',
                                                   description=f'API:\n{self.report_desc}')
            runner.run(suit)
        except IOError as ex:
            self.logger.error(str(ex))
        finally:
            self.logger.info("********TEST END*********")
            fp.close()


class BRReporter(Reporter):
    """BeautifulReport"""

    def __init__(self, case_path, report_theme=None,  *args, **kwargs):
        super().__init__(case_path, *args, **kwargs)
        self.report_theme = report_theme if report_theme else self.report.theme

    def run(self):
        if not os.path.exists(self.report_dir):
            os.makedirs(self.report_dir)
        suit = self.load_tests()
        result = BeautifulReport(suit)
        result.report(
            filename=self.report_file_name,
            description=self.report_title,
            report_dir=self.report_dir,
            theme=self.report_theme
        )



from os.path import dirname, join
from setuptools import setup, find_packages


with open(join(dirname(__file__), 'api_test_ez/VERSION'), 'rb') as f:
    version = f.read().decode('ascii').strip()


install_requires = [
    'tablib==3.2.1',
    'ddt~=1.5.0',
    'urllib3~=1.26.9',
    'requests~=2.28.0',
    'pytest~=7.1.2',
    'openpyxl==3.0.10',
    'BeautifulReport~=0.1.3',
]


setup(
    name='ApiTestEz',
    version=version,
    url='https://github.com/bruce4520196/ApiTestEz',
    description='An easier api test framework.',
    author='Bruce Cai',
    maintainer='Bruce Cai',
    maintainer_email='whiteghostcat@gmail.com',
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Framework :: Scrapy',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.5',
    install_requires=install_requires,
)
